/*
 * Copyright (c) 2012, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef MYVIZ_H
#define MYVIZ_H



#include <QWidget>
#include <QString>
#include <QGridLayout>
#include <ros/ros.h>
#include "std_msgs/String.h"
#include <thread>
#include <visualization_msgs/Marker.h>

#include <QTimer>
#include <QMutex>
#include <QProcess>
#include <QThread>
#include <QSlider>
#include <rosbag/bag.h>
#include <sensor_msgs/LaserScan.h>
#include <sensor_msgs/PointCloud2.h>

#include <rosbag/view.h>
#include"loginDialog.h"

namespace rviz
{
class Display;
class RenderPanel;
class VisualizationManager;
class ToolManager;
class PropertyTreeWidget;
class ViewManager;
class ViewController;
}

// BEGIN_TUTORIAL
// Class "MyViz" implements the top level widget for this example.
class MyViz: public QWidget
{
Q_OBJECT
public:
  MyViz( QWidget* parent = 0 );
  virtual ~MyViz();
  void subCallback(const std_msgs::String& msg);
  void pubThread();
  void PlayCircles();  // 圆圈线程执行函数
  rosbag::View *view;
  rosbag::View *view_;

private Q_SLOTS:   // QT信号和槽
  void setThickness( int thickness_percent );
  void setCellSize( int cell_size_percent );
  void setCloudTopic(const QString &newTopic);
  void setCloudSize(const QString& size);
  void setCloudAlpha(const QString& text);
  void setLaserAlpha(const QString& laserAlpha);

  void setLaserTopic(const QString &newTopic);
  void setLaserSize(const QString & laserSize);
  void LaserColorChanged(int index); // 雷达颜色
  void GridColorChanged(int index); // 格栅颜色
  void CloudColorChanged(int index);
  void setFixedFrame(const QString &FixedFrame);
  void setBackgroundColor(const QString& bg_color); // 背景颜色
  void AxesDisplayChanged(int index); // axe
  void Tree_Display(QGridLayout* layout, int index);

  void login_button_clicked();
  void quit_login_button_clicked();
  void slot_select(); // 切换至select模式
  void slot_move_camera();
  void CircleDisplayAdded(); // 增加圆圈按钮
  void CircleDisplayDeleted(); // 删除圆圈按钮
  void ViewZero(); // 视角回正
  void ViewLeft(); // 视角左转
  void ViewRight(); // 视角右转
  void ViewTop(); // 视角俯视

  // off
  double extractDuration(const QByteArray& output);
  void selectFile();
  void SpeedChanged(int index);
  void openFile();
  void pcdFile();
  void plyFile();
  void bin();
  std::string find(const std::string Path);

  void pcd2ply (std::string &in_file, std::string& out_file);
  void pcd2bin (std::string &in_file, std::string& out_file);
  void ply2pcd (std::string &in_file, std::string& out_file);
  void ply2bin (std::string &in_file, std::string& out_file);
  void bin2pcd (std::string &in_file, std::string& out_file);
  void bin2ply (std::string &in_file, std::string& out_file);
  void TransformeType();

private:
  rviz::VisualizationManager* manager_; // 
  rviz::RenderPanel* render_panel_;
  rviz::ToolManager* tool_manager_;
  rviz::ViewManager* view_man_; // 视角总类
  rviz::ViewController* viewController; // 视角控制类
  rviz::Display* grid_;
  rviz::Display* cloud;
  rviz::Display* laser_;
  rviz::Display* axes_display;
  rviz::Display* marker_;
  rviz::PropertyTreeWidget* select_tree_widget_ ; // select框

  // rviz::PropertyTreeWidget* view_tree_widget_ ;
  ros::NodeHandle nh;
  ros::Subscriber sub;
  ros::Publisher pub;
  std::thread pub_thread;

  std::vector<QString> color_name; // 颜色的名称集合
  // std::vector<QString> button_name; // 带图标按钮的名称集合
  QString color_icon_filename;
  QString button_icon_filename;
  
  LoginDialog* dlg;
  bool landed;
  QString cur_fixed_frame;
  int circle_index;
  std::map<double, int> circle_radius;
  // off
  bool ispauseclick;// 用于判断暂停键第几次按下
  int count;// 用于记录timeout的次数
  bool sliderreleased;// 用于判断是否拖动了进度条
  bool playagain;// 用于检测是否多次按下play按钮
  bool isthreadplaying;// 用于标识线程是否在运行
  bool stored = false;// 用于存储一次容器,只在第一次play的时候存储容器
  bool is_pause = false;
  bool is_reset = false;
  bool is_end = false;
  bool is_save = false;
  bool is_saveMul = false;
  bool is_bagClosed = true;
  bool is_rateChanged = false;
  bool is_openFile = false;
  bool is_setFrameIndex = false;
  bool is_saveMulFinish = false;
  bool end = false;
  bool threadout = false;
  bool ischanged = false;
  bool ismoved = false;
  
  bool sliderpause = false;
  bool is_next = false;
  bool is_previous = false;
  bool flag = false;
  bool pointcloud2 = false;
  double speed_rate = 10;
  QPushButton* play;
  QSlider* progress;
  QString filePath;
  QString Topic;
  QString SavefileName;
  QString SaveMulfileName;

  double sliderPosition = 0;
  int prog_num = 0;
  std::vector<QString> speed_name;
  ros::Publisher scan_pub;
  ros::Publisher scan_pubpcd;
  std::vector<sensor_msgs::LaserScan::ConstPtr> indx;// 存储laser节点信息
  std::vector<sensor_msgs::PointCloud2::ConstPtr> indxpcd;// 存储cloud节点信息
  int countbag = 0;// 存储播放的具体位置
  double indxsize = 0;
  double indxpcdsize = 0;
  int test = 0;


};
// END_TUTORIAL
#endif // MYVIZ_H
